# -*- coding: utf-8 -*-
from optimus.utils.assets import synchronize_assets_sources
from optimus.utils.directory import (
    init_directory,
    recursive_directories_create,
    initialize,
)
from optimus.utils.settings import display_settings
from optimus.utils.server import get_host_parts


__all__ = [
    "synchronize_assets_sources",
    "init_directory",
    "recursive_directories_create",
    "initialize",
    "display_settings",
    "get_host_parts",
]



# -*- coding: utf-8 -*-
"""
Exceptions
==========

Specific exceptions that Optimus code can raise.
"""


class OptimusBaseException(Exception):
    """
    Base for Optimus exceptions.
    """

    pass


class DestinationExists(OptimusBaseException):
    """
    Exception to be raised when a destination allready exists for a new project
    to create.
    """

    pass


class SettingsImportError(OptimusBaseException):
    """
    Exception to be raised when a settings module import fails.
    """

    pass


class TemplateImportError(OptimusBaseException):
    """
    Exception to be raised when a template module import fails.
    """

    pass


class TemplateSettingsInvalidError(OptimusBaseException):
    """
    Exception to be raised when a template manifest have wrong settings.
    """

    pass


class InvalidLangageIdentifier(OptimusBaseException):
    """
    Exception to be raised from ``lang.LangBase`` when given language code is
    invalid.
    """

    pass


class InvalidSettings(OptimusBaseException):
    """
    Exception to be raised for invalid settings from 'conf.model'
    """

    pass


class ViewImproperlyConfigured(OptimusBaseException):
    """
    Exception to be raised from ``pages.view.base.PageViewBase`` when
    instanciated with bad value or missing parameters.
    """

    pass


class ServerConfigurationError(OptimusBaseException):
    """
    Exception to be raised when trying to run live server with a error in given
    configuration.
    """

    pass


class InvalidHostname(ServerConfigurationError):
    """
    Exception to be raised when a parsed hostname is invalid.
    """

    pass



# -*- coding: utf-8 -*-
import os
import logging

from optimus.exceptions import ServerConfigurationError
from optimus.utils import get_host_parts


try:
    import cherrypy
except ImportError:
    CHERRYPY_AVAILABLE = False
else:
    CHERRYPY_AVAILABLE = True


def server_interface(settings, hostname, index="index.html"):
    """
    Validate configuration, configurate cherrypy server and application to serve build
    directory.

    Arguments:
        settings (optimus.conf.model.SettingsModel): Settings object which defines
            everything required for building.
        hostname (string): Hostname required to bind on. This is the host address with
            optional port like ``localhost:8001`` or without port ``localhost``
            (default HTTP port ``80`` will be used).

    Keyword Arguments:
        index (string): Filename to consider as directory index. Default to
            ``index.html``.

    Returns:
        dict: A dictionnary with cherrypy module object (``cherrypy`` item), a dict
        for application config (``app_conf`` item) and the relative URL to mount
        application (``mount_on`` item) which is usally the root.
    """
    logger = logging.getLogger("optimus")

    if not CHERRYPY_AVAILABLE:
        msg = (
            "Unable to import CherryPy, you need to install it with 'pip install "
            "cherrypy' in your environment."
        )
        raise ServerConfigurationError(msg)

    # Parse hostname to distinct address and port
    address, port = get_host_parts(hostname)

    # Check project publish directory exists
    if not os.path.exists(settings.PUBLISH_DIR):
        msg = "Publish directory does not exist yet, you need to build it before: {}"
        raise ServerConfigurationError(msg.format(settings.PUBLISH_DIR))

    # Run server with publish directory served with tools.staticdir
    msg = "Running HTTP server on address '{address}' with port '{port}'"
    logger.info(msg.format(address=address, port=port))

    # Configure webapp server
    cherrypy.config.update(
        {
            "server.socket_host": address,
            "server.socket_port": port,
            "engine.autoreload_on": False,
        }
    )

    # Configure webapp for static files
    app_conf = {
        "/": {
            "tools.staticdir.index": index,
            "tools.staticdir.on": True,
            "tools.staticdir.dir": settings.PUBLISH_DIR,
        },
    }

    return {
        "cherrypy": cherrypy,
        "app_conf": app_conf,
        "mount_on": "/",
    }



# -*- coding: utf-8 -*-
import importlib
import os

import click

from optimus.conf.loader import import_settings_module, load_settings
from optimus.interfaces.po import po_interface
from optimus.setup_project import setup_project
from optimus.utils import display_settings


@click.command("po", short_help="Manage project translation catalogs")
@click.option(
    "--init",
    is_flag=True,
    help=(
        "Initialize structure, create template catalog (POT) and "
        "initialize catalogs (PO)"
    ),
)
@click.option(
    "--update",
    is_flag=True,
    help=(
        "Extract translations, update the template catalog (POT) "
        "and update the catalogs (PO)"
    ),
)
@click.option(
    "--compile",
    "compile_opt",
    is_flag=True,
    help=("Process to compilation of catalogs"),
)
@click.option(
    "--basedir",
    metavar="PATH",
    type=click.Path(exists=True),
    help=(
        "Base directory where to search for settings file. "
        "Default value use current directory."
    ),
    default=os.getcwd(),
)
@click.option(
    "--settings-name",
    metavar="NAME",
    help=(
        "Settings file name to use without '.py' extension. "
        "Default value is 'settings'."
    ),
    default="settings",
)
@click.pass_context
def po_command(context, init, update, compile_opt, basedir, settings_name):
    """
    Manage project translation catalogs for all registred languages
    """
    # Set project before to be able to load its modules
    setup_project(basedir, settings_name)

    # Load current project settings and page map
    settings = import_settings_module(settings_name, basedir=basedir)
    # In test environment, force the module reload to avoid previous test cache to be
    # used (since the module have the same path).
    if context.obj["test_env"]:
        settings = importlib.reload(settings)

    settings = load_settings(settings)

    # Debug output
    display_settings(
        settings,
        ("DEBUG", "PROJECT_DIR", "SOURCES_DIR", "TEMPLATES_DIR", "LOCALES_DIR"),
    )

    po_interface(settings, init=init, update=update, compile_opt=compile_opt)



# -*- coding: utf-8 -*-
from watchdog.observers import Observer

from optimus.watchers.templates import TemplatesWatchEventHandler
from optimus.watchers.assets import AssetsWatchEventHandler


def watcher_interface(settings, views, build_env):
    """
    Initialize observer for views and possible assets according to settings and build
    environment.

    Commonly before using this function you will use ``builder_interface`` first since
    it will perform a first (required) build and init the builder environment as
    expected in ``build_env`` argument.

    Once this interface returns the observer object, you may use it like so: ::

        observer.start()
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()

    Arguments:
        settings (optimus.conf.model.SettingsModel): Settings object which defines
            everything required for building.
        views (object): Module which defines page views to build, in fact the module
            object require only a ``PAGES`` attribute that is a list of Page view.
        build_env (dict): A dictionnary with initialized builder (``builder`` item),
            asset manager (``assets_env`` item) and the list of builded pages
            (``builded`` item).

    Returns:
        watchdog.observers.Observer: The initialized and configured observer for
        setted watchers.
    """
    # Perform a first scanning of page templates
    build_env["builder"].scan_bulk(views.PAGES)

    # Init templates events watchers
    templates_event_handler = TemplatesWatchEventHandler(
        settings, build_env["builder"], **settings.WATCHER_TEMPLATES_PATTERNS
    )

    # Init assets events watchers
    if build_env["assets_env"] is not None:
        assets_event_handler = AssetsWatchEventHandler(
            settings,
            build_env["assets_env"],
            build_env["builder"],
            **settings.WATCHER_ASSETS_PATTERNS,
        )

    # Initialize observer to use
    observer = Observer()

    # Register views events watcher
    observer.schedule(
        templates_event_handler,
        settings.TEMPLATES_DIR,
        recursive=True,
    )
    # Register assets events watcher
    # NOTE: This observe a single handler for every assets for simplicity. However the
    #       drawback is it watch on the whole sources directory, but it has been
    #       configured to care only about assets (css and js files) so it may not be
    #       a performance leak except on very huge source directory. The only issue is
    #       that it may trigger some event for css or js files which are in templates.
    if build_env["assets_env"] is not None:
        observer.schedule(
            assets_event_handler,
            settings.SOURCES_DIR,
            recursive=True,
        )

    return observer



# -*- coding: utf-8 -*-
import importlib
import logging
import os

import click

from optimus.conf.loader import import_settings_module, load_settings
from optimus.exceptions import ServerConfigurationError
from optimus.setup_project import setup_project
from optimus.utils import display_settings
from optimus.interfaces.runserver import server_interface


# TODO: Change settings-name to settings ?
@click.command(
    "runserver", short_help=("Launch a simple HTTP server on " "built project")
)
@click.argument("hostname", default="127.0.0.1:80")
@click.option(
    "--basedir",
    metavar="PATH",
    type=click.Path(exists=True),
    help=(
        "Base directory where to search for settings file. "
        "Default value use current directory."
    ),
    default=os.getcwd(),
)
@click.option(
    "--settings-name",
    metavar="NAME",
    help=(
        "Settings file name to use without '.py' extension. "
        "Default value is 'settings'."
    ),
    default="settings",
)
@click.option(
    "--index",
    metavar="FILENAME",
    help=("Filename to use as directory index. " "Default value is 'index.html'."),
    default="index.html",
)
@click.pass_context
def runserver_command(context, basedir, settings_name, index, hostname):
    """
    Launch a simple HTTP server rooted on the project build directory

    Default behavior is to bind server on IP address '127.0.0.1' and port
    '80'. You may give another host to bind to as argument 'HOSTNAME'.

    'HOSTNAME' can be either a simple address like '0.0.0.0' or an address and
    port like '0.0.0.0:8001'. If no custom port is given, '80' is used as
    default.
    """
    logger = logging.getLogger("optimus")

    # Set project before to be able to load its modules
    setup_project(basedir, settings_name)

    # Load current project settings and page map
    settings = import_settings_module(settings_name, basedir=basedir)
    # In test environment, force the module reload to avoid previous test cache to be
    # used (since the module have the same path).
    if context.obj["test_env"]:
        settings = importlib.reload(settings)

    settings = load_settings(settings)

    # Debug output
    display_settings(
        settings,
        ("DEBUG", "PROJECT_DIR", "SOURCES_DIR", "TEMPLATES_DIR", "LOCALES_DIR"),
    )

    try:
        server_env = server_interface(settings, hostname, index=index)
    except ServerConfigurationError as e:
        logger.error(e)
        raise click.Abort()

    # Don't start cherrypy server during tests
    if not context.obj["test_env"]:
        server_env["cherrypy"].quickstart(
            None,
            server_env["mount_on"],
            config=server_env["app_conf"],
        )



# -*- coding: utf-8 -*-
"""
Settings model
**************

"""
import os

from optimus.exceptions import InvalidSettings


class SettingsModel(object):
    """
    Settings model

    Basically empty on init, you'll have to fill it either from kwargs or a
    module (or an object).

    Be aware that on default behavior settings loading methods execute
    ``apply_defaults`` method that will apply default values for some settings
    but related settings are not updated, you need to take care of them
    yourself or disabled default values pratice.

    Attributes:
        _excluded_names (list): Setting names to exclude from loading methods.
        _required_settings (list): Settings names required to be defined from
            loading methods.
    """

    _excluded_names = []
    _required_settings = (
        "PROJECT_DIR",
        "DEBUG",
        "SITE_NAME",
        "SITE_DOMAIN",
        "SOURCES_DIR",
        "TEMPLATES_DIR",
        "PUBLISH_DIR",
        "STATIC_DIR",
        "STATIC_URL",
    )

    def validate_name(self, name):
        """
        Filter to validate setting name

        Name must be uppercase, not starting with a '_' character and not
        registred in exluded names.

        Arguments:
            name (string): Setting name.

        Returns:
            bool: True if name is valid, else False.
        """
        return (
            name not in self._excluded_names
            and not name.startswith("_")
            and name.isupper()
        )

    def check(self):
        """
        Check every required settings are defined
        """
        missing_settings = []

        for setting_name in self._required_settings:
            if not hasattr(self, setting_name):
                missing_settings.append(setting_name)

        if len(missing_settings) > 0:
            msg = (
                "The following settings are required but not " "defined: {0}"
            ).format(", ".join(missing_settings))
            raise InvalidSettings(msg)

    def _default_jinja(self):
        """
        Set default attributes for required settings around Jinja
        """
        # Python paths for each extensions to use with Jinja2
        if not hasattr(self, "JINJA_EXTENSIONS"):
            self.JINJA_EXTENSIONS = ("jinja2.ext.i18n",)

        # Template filters to use with Jinja2
        if not hasattr(self, "JINJA_FILTERS"):
            self.JINJA_FILTERS = {}

    def _default_watchdog(self):
        """
        Set default attributes for required settings Watchdog
        """
        # Templates watcher settings
        if not hasattr(self, "WATCHER_TEMPLATES_PATTERNS"):
            self.WATCHER_TEMPLATES_PATTERNS = {
                "patterns": ["*.html"],
                "ignore_patterns": None,
                "ignore_directories": False,
                "case_sensitive": False,
            }
        # Assets watcher settings
        if not hasattr(self, "WATCHER_ASSETS_PATTERNS"):
            self.WATCHER_ASSETS_PATTERNS = {
                "patterns": ["*.css", "*.js", "*.json"],
                "ignore_patterns": None,
                "ignore_directories": False,
                "case_sensitive": False,
            }

    def _default_webassets(self):
        """
        Set default attributes for required settings around Webassets
        """
        # The directory where webassets will store his cache
        if not hasattr(self, "WEBASSETS_CACHE"):
            self.WEBASSETS_CACHE = os.path.join(self.PROJECT_DIR, ".webassets-cache")

        # Asset version appended through url querystring behavior
        if not hasattr(self, "WEBASSETS_URLEXPIRE"):
            self.WEBASSETS_URLEXPIRE = None

        # Bundles
        if not hasattr(self, "BUNDLES"):
            self.BUNDLES = {}

        if not hasattr(self, "ENABLED_BUNDLES"):
            self.ENABLED_BUNDLES = list(self.BUNDLES.keys())

    def _default_babel(self):
        """
        Set default attributes for required settings around Babel
        """
        # Default directory for translation catalog
        if not hasattr(self, "LOCALES_DIR"):
            self.LOCALES_DIR = os.path.join(self.PROJECT_DIR, "locale")

        # Default used language
        if not hasattr(self, "LANGUAGE_CODE"):
            self.LANGUAGE_CODE = "en_US"

        # Default available languages to manage
        if not hasattr(self, "LANGUAGES"):
            self.LANGUAGES = (self.LANGUAGE_CODE,)

        # Default map for translaction extract with babel
        if not hasattr(self, "I18N_EXTRACT_MAP"):
            self.I18N_EXTRACT_MAP = (
                ("pages.py", "python"),
                ("*settings.py", "python"),
                ("**/templates/**.html", "jinja2"),
            )

        if not hasattr(self, "I18N_EXTRACT_OPTIONS"):
            self.I18N_EXTRACT_OPTIONS = {
                "**/templates/**.html": {
                    "extensions": "webassets.ext.jinja2.AssetsExtension",
                    "encoding": "utf-8",
                },
            }

        if not hasattr(self, "I18N_EXTRACT_SOURCES"):
            self.I18N_EXTRACT_SOURCES = (self.PROJECT_DIR,)

    def apply_defaults(self):
        """
        Apply default attributes for needed but not required settings.
        """
        # Python path to the file that contains pages map, this is relative to
        # project directory
        if not hasattr(self, "PAGES_MAP"):
            self.PAGES_MAP = "pages"

        # Directories to "synchronize" within static directory
        if not hasattr(self, "FILES_TO_SYNC"):
            self.FILES_TO_SYNC = ()

        # Use https protocol instead of http if enabled
        if not hasattr(self, "HTTPS_ENABLED"):
            self.HTTPS_ENABLED = False

        # Default settings for dependancies
        self._default_watchdog()
        self._default_jinja()
        self._default_webassets()
        self._default_babel()

    def load_from_kwargs(self, check=True, defaults=True, **kwargs):
        """
        Set setting attribute from given named arguments.

        Keyword Arguments:
            check (bool): True to perform required settings check. Default is
                True.
            defaults (bool): True to set default needed settings.
            kwargs: Named arguments to load as settings.

        Returns:
            list: List of loaded setting names from given arguments.
        """
        setted = []

        for name in kwargs:
            if self.validate_name(name):
                setattr(self, name, kwargs.get(name))
                setted.append(name)

        if check:
            self.check()
        if defaults:
            self.apply_defaults()

        return setted

    def load_from_module(self, settings_module, check=True, defaults=True):
        """
        Set setting attribute from given module variables.

        Keyword Arguments:
            settings_module (object): Object to find attributes to load as
                settings. Every valid attribute names will be used.
            defaults (bool): True to set default needed settings.
            kwargs: Named arguments to load as settings.

        Returns:
            list: List of loaded setting names from given module.
        """
        setted = []

        for name in dir(settings_module):
            if self.validate_name(name):
                setattr(self, name, getattr(settings_module, name))
                setted.append(name)

        if check:
            self.check()
        if defaults:
            self.apply_defaults()

        return setted



import logging
import os
import sys

from optimus import PROJECT_DIR_ENVVAR, SETTINGS_NAME_ENVVAR


def setup_project(basedir, settings_name, set_envvar=True, set_syspath=True):
    """
    Setup for an Optimus project.

    Arguments:
        basedir (string): Absolute path to project base directory, aka the directory
            which contains the settings and pages modules (can be in subdirectories
            if they are correct module directories). This path must exists and be a
            valid directory.
        settings_name (string): Default settings module name to set in environment
            variable.

    Keyword Arguments:
        set_envvar (boolean): Enable the os environment variable setup. Default is True.
        set_syspath (boolean): Enable adding base directory path to ``sys.path``.
            Default is True.
    """
    logger = logging.getLogger("optimus")
    status = {}

    # Check given base directory
    if not os.path.exists(basedir):
        msg = "Given base directory path does not exists: {}"
        raise ImportError(msg.format(basedir))
    elif not os.path.isdir(basedir):
        msg = "Given base directory path is not a directory: {}"
        raise ImportError(msg.format(basedir))

    # Set required environment variables to load settings
    if set_envvar:
        if PROJECT_DIR_ENVVAR not in os.environ or not os.environ[PROJECT_DIR_ENVVAR]:
            os.environ[PROJECT_DIR_ENVVAR] = basedir
            status["set_os_environ_project"] = basedir
        if (
            SETTINGS_NAME_ENVVAR not in os.environ
            or not os.environ[SETTINGS_NAME_ENVVAR]
        ):
            os.environ[SETTINGS_NAME_ENVVAR] = settings_name
            status["set_os_environ_settings"] = settings_name

    # Add project base directory to the sys.path so its modules can be imported easily
    if set_syspath and basedir not in sys.path:
        logger.info("Register project base directory: {}".format(basedir))
        sys.path.append(basedir)
        status["add_basedir_to_syspath"] = basedir

    return status



# -*- coding: utf-8 -*-
"""
.. _watchdog: https://github.com/gorakhargosh/watchdog

Watcher handlers
================

This component contains `watchdog`_ handlers to use to watch a project
sources.

"""
import os


class BaseHandler(object):
    """
    Base class for handlers.

    Assume children inheriting this class have a ``settings`` attribute with a
    valid ``SettingsModel`` instance as value.
    """

    def get_relative_template_path(self, path):
        """
        Retrieve relative path from templates directory.

        Arguments:
            path (str): Path to a template file.

        Returns:
            string: Relative path either from templates directory or untouched
            if file does not belong to template dir.
        """
        if path.startswith(self.settings.TEMPLATES_DIR):
            return os.path.relpath(path, self.settings.TEMPLATES_DIR)

        return path

    def get_relative_asset_path(self, path):
        """
        Retrieve relative path from assets directory.

        Arguments:
            path (str): Path to an asset file.

        Returns:
            string: Relative path either from assets directory or untouched
            if file does not belong to assets dir.
        """
        if path.startswith(self.settings.SOURCES_DIR):
            return os.path.relpath(path, self.settings.SOURCES_DIR)

        return path



# -*- coding: utf-8 -*-
"""
Settings registration
*********************

This is the entry point to reach settings from project page module.
"""
import os

from optimus import PROJECT_DIR_ENVVAR, SETTINGS_NAME_ENVVAR
from optimus.conf.loader import import_settings

try:
    basedir = os.environ[PROJECT_DIR_ENVVAR]
    # If it's set but is an empty string.
    if not basedir:
        raise KeyError
except KeyError:
    raise ImportError(
        (
            "Project cannot be imported, because "
            "environment variable {} is "
            "undefined."
        ).format(PROJECT_DIR_ENVVAR)
    )

try:
    name = os.environ[SETTINGS_NAME_ENVVAR]
    # If it's set but is an empty string.
    if not name:
        raise KeyError
except KeyError:
    raise ImportError(
        (
            "Settings cannot be imported, because "
            "environment variable {} is "
            "undefined."
        ).format(SETTINGS_NAME_ENVVAR)
    )

# Reachable settings object
settings = import_settings(name, basedir)



# -*- coding: utf-8 -*-
import logging

from watchdog.events import PatternMatchingEventHandler
from watchdog.utils.patterns import match_any_paths

from optimus.watchers import BaseHandler


class TemplatesWatchEventHandler(BaseHandler, PatternMatchingEventHandler):
    """
    Template events handler.

    Arguments:
        settings (optimus.conf.model.SettingsModel): Project settings.
        pages_builder (optimus.pages.builder.PageBuilder): Page builder object
            that is triggered to perform page building.
        args: Additional arguments to be passed to handler, commonly for
            watchdog API.

    Keyword Arguments:
        kwargs: Optionnal keyword arguments commonly for watchdog API.

    Attributes:
        settings (optimus.conf.model.SettingsModel): As given from arguments.
        pages_builder (optimus.pages.builder.PageBuilder): As given from
            arguments.
        logger (logging.Logger): Optimus logger.
    """

    def __init__(self, settings, pages_builder, *args, **kwargs):
        self.settings = settings
        self.pages_builder = pages_builder

        self.logger = logging.getLogger("optimus")

        super(TemplatesWatchEventHandler, self).__init__(*args, **kwargs)

    def build_for_item(self, path):
        """
        Build all pages using given template path.

        If template is a snippet included in other templates they will be
        flagged for build too. This is recursive so a snippet in a snippet in
        a snippet will raises to page templates.

        ``path`` argument is a template path

        Arguments:
            path (string): Template path.

        Returns:
            list: List of builded pages.
        """
        rel_path = self.get_relative_template_path(path)
        built = []
        # Search in the registry if the file is a knowed template dependancy
        if rel_path in self.pages_builder.registry.elements:
            msg = "--- Changes detected on: {} ---"
            self.logger.warning(msg.format(rel_path))

            requires = self.pages_builder.registry.get_pages_from_dependency(
                rel_path
            )  # noqa

            self.logger.debug("Requires for rebuild: {}".format(requires))

            builds = self.pages_builder.build_bulk(requires)
            built.extend(builds)

        return built

    def on_moved(self, event):
        """
        Called when a file or a directory is moved or renamed.

        Many editors don't directly change a file, instead they make a
        transitional file like ``*.part`` then move it to the final filename.

        Arguments:
            event: Watchdog event, either ``watchdog.events.DirMovedEvent`` or
                ``watchdog.events.FileModifiedEvent``.
        """
        # We are only interested for the destination
        if match_any_paths(
            [event.dest_path],
            included_patterns=self.patterns,
            excluded_patterns=self.ignore_patterns,
            case_sensitive=self.case_sensitive,
        ):
            msg = "Change detected from a move on: {}"
            self.logger.debug(msg.format(event.dest_path))

            return self.build_for_item(event.dest_path)

        return []

    def on_created(self, event):
        """
        Called when a new file or directory is created.

        Arguments:
            event: Watchdog event, either ``watchdog.events.DirCreatedEvent``
                or ``watchdog.events.FileCreatedEvent``.
        """
        msg = "Change detected from a create on: {}"
        self.logger.debug(msg.format(event.src_path))

        return self.build_for_item(event.src_path)

    def on_modified(self, event):
        """
        Called when a file or directory is modified.

        Arguments:
            event: Watchdog event, ``watchdog.events.DirModifiedEvent`` or
                ``watchdog.events.FileModifiedEvent``.
        """
        msg = "Change detected from an edit on: {}"
        self.logger.debug(msg.format(event.src_path))

        return self.build_for_item(event.src_path)



# -*- coding: utf-8 -*-
"""
Langage base object
********************

"""
from optimus.exceptions import InvalidLangageIdentifier


class LangBase:
    """
    Langage base object to encapsulate the language label, code and other
    details.

    Alternative and External code are not really used internally in optimus,
    there are only for some template usage.

    The instance will also supply a "language_name" and "region_name" class
    attributes, which are the result of splitting the code on two parts.
    "region_name" is ``None`` by default, as the region name is optional in
    language identifier.

    Code and name is not validated, you are responsible to ensure they are
    valid.

    See http://www.i18nguy.com/unicode/language-identifiers.html for more
    details on language identifiers.

    Example: ::

        class LangFr(LangBase):
            code = 'fr'
            label = 'France'

    Or: ::

        lang = LangBase(code="zh_CN", label="Chinese")

    Keyword Arguments:
        code (string): Langage identifier.
        label (string): Langage label like "Français" for ``fr``.

    Attributes:
        label (string): Default language label if not given in kwargs.
        code (string): Default language identifier if not given in kwargs.
        alt_code (string): Alternative code, will be equal to "code" if not
            set.
        external_code (string): External code for some external apps, will be
            equal to ``alt_code`` if not set.
    """

    label = None
    code = None
    alt_code = None
    external_code = None

    def __init__(self, code=None, label=None):
        self.code = code or self.code

        if self.code is None:
            msg = (
                """Missing language identifier : You must supply it by """
                """the way of 'code' argument or as the 'code' class """
                """attribute."""
            )
            raise InvalidLangageIdentifier(msg)

        if len(self.code.split("-")) > 1:
            msg = (
                """Invalid language identifier : Langage name and """
                """region name must be joined by a '_' not a '-'"""
            )
            raise InvalidLangageIdentifier(msg)

        self.language_name, self.region_name = self.split_code(self.code)

        self.label = label or self.label or self.code

        self.alt_code = self.alt_code or self.code
        self.external_code = self.external_code or self.code

    def __str__(self):
        return self.label

    def __repr__(self):
        """
        Object representation

        Returns:
            string: Representation with name and code
        """
        return "<{name} code:{code}>".format(
            name=self.__class__.__name__, code=self.code
        )

    def split_code(self, code):
        """
        Split language identifier to language name and region name (if any).

        Arguments:
            code (string): Langage identifier.

        Returns:
            tuple: A pair of language name and possibly region name, if code
            does not contain any region it will be ``None``.
        """
        items = code.split("_")
        language_name = items[0]
        region_name = None

        if len(items) > 1:
            region_name = items[1]

        return language_name, region_name



# -*- coding: utf-8 -*-
import logging


def display_settings(settings, names):
    """
    Helper to output values of given setting names to logger.

    Arguments:
        settings (object): Settings object.
        names (list): List of setting name to output. If a name item does not
            exists as attribute in given ``settings`` object, its value will
            be ``NOT SET``.

    """
    logger = logging.getLogger("optimus")
    for item in names:
        value = getattr(settings, item, "NOT SET")
        logger.debug(" - Settings.{} = {}".format(item, value))



# -*- coding: utf-8 -*-
import logging
import os

from jinja2 import meta as Jinja2Meta

from optimus.exceptions import ViewImproperlyConfigured
from optimus.i18n.lang import LangBase


class PageViewBase:
    """
    Base view object for a page

    You can set class attributes at the init if needed

    The render method is responsible to rendering the HTML from the template
    and his context. Actually this is the only used method directly.

    Only ``lang`` and ``context`` attributes are optional, so take care to set
    all the required ones because their default value is ``None``. You should
    not use directly ``PageViewBase``, inherit it in a common object with all
    attributes setted by default.

    Template context will have the following variables :

    page_title
        Page title
    page_destination
        Page destination
    page_lang
        Given langage if any
    page_template_name
        Template name used to compile the page HTML

    But you can add new variable if needed. The default context variables can
    not be overriden from the ``context`` class attribute, only from the
    ``get_context`` class method.

    View need settings to be defined either as argument on instance init or
    later through attribute setter.

    Attributes:
        title (string): Page title.
        template_name (string): Page template file path relaive to templates
            directoy. Used as Python template string with optional non
            positional argument ``{{ language_code }}`` available for
            internationalized pages.
        destination (string): Page destionation path relative to build
            directory.
        lang (string): Language identifier or an instance of
            ``optimus.i18n.LangBase``.
        context (dict): Initial page template context.
        logger (logging.Logger): Optimus logger.
        _used_templates (list): List of every used templates. Only filled when
            ``introspect()`` method is executed. Default to ``None``.
        __settings (conf.model.SettingsModel): Settings registry instance when
            given in kwargs. Default to ``None``.

    Arguments:
        **kwargs: Arbitrary keyword arguments. Will be added as object
            attribute.
    """

    title = None
    template_name = None
    destination = None
    lang = None
    context = {}

    def __init__(self, **kwargs):
        self._used_templates = None
        self.logger = logging.getLogger("optimus")
        self.__settings = kwargs.pop("settings", None)

        # Store every passed keyword argument as object attribute
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.validate()

    def __str__(self):
        return self.get_destination()

    def __repr__(self):
        """
        Object representation

        Returns:
            string: Representation with name and code
        """
        return "<{name} {dest}>".format(
            name=self.__class__.__name__, dest=self.get_destination()
        )

    def validate(self):
        """
        Validate every required attribute is set.

        Returns:
            boolean: ``True`` if requirements are set.
        """
        err = []

        for item in ["title", "template_name", "destination"]:
            if not getattr(self, item):
                err.append(item)

        if len(err) > 0:
            msg = "These attributes are required: {}".format(", ".join(err))
            raise ViewImproperlyConfigured(msg)

        return True

    @property
    def settings(self):
        """
        ``settings`` attribute getter, check settings have been correctly
        defined.

        Returns:
            conf.model.SettingsModel: Settings registry instance when given
            in kwargs. Default to ``None``.
        """
        if not self.__settings:
            msg = (
                """View required settings defined either from init """
                """arguments or through settings attribute"""
            )
            raise ViewImproperlyConfigured(msg)
        return self.__settings

    @settings.setter
    def settings(self, settings):
        """
        ``settings`` attribute setter

        Arguments:
            settings (conf.model.SettingsModel): Settings registry instance.
        """
        self.__settings = settings

    def get_title(self):
        """
        Get page title.

        Default behavior is to used page attribute ``title``.

        Returns:
            string: Page title.
        """
        return self.title

    def get_lang(self):
        """
        Get page language object.

        Returns:
            optimus.i18n.LangBase: Language object. If ``lang`` page attribute
            is ``None`` it will create a language object using default
            language identifier from setting ``LANGUAGE_CODE``.
        """
        # Defaut language identifier if not given
        if getattr(self, "lang", None) is None:
            self.lang = LangBase(code=self.settings.LANGUAGE_CODE)
        # If the lang attribute contains a string, assume this is the language
        # identifier
        elif isinstance(getattr(self, "lang"), str):
            self.lang = LangBase(code=getattr(self, "lang"))

        return self.lang

    def get_destination(self):
        """
        Get page destination path.

        Returns:
            string: Page destination path relative to build directory.
        """
        return os.path.normpath(
            self.destination.format(language_code=self.get_lang().code)
        )

    def get_relative_position(self):
        """
        Get relative path position from the destination file to the root.

        Returns:
            string: Either something like "../../" if the destination is in
            subdirectories or "./" if at the root. Won't never return empty
            string.
        """
        return (len(self.get_destination().split("/")) - 1) * "../" or "./"

    def get_template_name(self):
        """
        Get template file path.

        Returns:
            string: Template file path relative to templates directory.
        """
        return self.template_name.format(language_code=self.get_lang().code)

    def get_context(self):
        """
        Get template context.

        Returns:
            dict: Template context of variables.
        """
        self.context.update(
            {
                "page_title": self.get_title(),
                "page_destination": self.get_destination(),
                "page_relative_position": self.get_relative_position(),
                "page_lang": self.get_lang(),
                "page_template_name": self.get_template_name(),
            }
        )

        self.logger.debug(" - Initial context: {}".format(self.context))

        return self.context

    def render(self, env):
        """
        Take the Jinja2 environment as required argument.

        Arguments:
            env (jinja2.Jinja2Environment): Jinja environment.

        Returns:
            string: HTML builded from page template with its context.
        """
        self.env = env
        context = self.get_context()

        template = self.env.get_template(self.get_template_name())

        return template.render(lang=self.get_lang(), **context)

    def _recurse_template_search(self, env, template_name):
        """
        Load involved template sources from given template file path then find
        their template references.

        Arguments:
            env (jinja2.Jinja2Environment): Jinja environment.
            template_name (string): Template file path.

        Returns:
            list: List of involved templates sources files.
        """
        template_source = env.loader.get_source(env, template_name)[0]
        parsed_content = env.parse(template_source)

        deps = []
        for item in Jinja2Meta.find_referenced_templates(parsed_content):
            deps.append(item)
            deps += self._recurse_template_search(env, item)

        return deps

    def introspect(self, env):
        """
        Take the Jinja2 environment as required argument to find every
        templates dependancies from page.

        Arguments:
            env (jinja2.Jinja2Environment): Jinja environment.

        Returns:
            list: List of involved templates sources files.
        """
        if self._used_templates is None:
            self.env = env

            found = self._recurse_template_search(env, self.get_template_name())

            self._used_templates = [self.get_template_name()] + found

            self.logger.debug(" - Used templates: {}".format(self._used_templates))

        return self._used_templates



# -*- coding: utf-8 -*-
import logging

from cookiecutter.main import cookiecutter
from cookiecutter.exceptions import RepositoryNotFound


def starter_interface(template, project_name, output_dir):
    """
    Interface to create a new Optimus project from a cookiecutter template.

    Be aware that cookiecutter emit a lot of logs, you may want to mute them.

    Raises:
        exceptions: Any possible exceptions from cookiecutter.

    Arguments:
        template (string): Path to a template directory or an URL to a public template
            repository.
        project_name (string): Project name. It would be used to make the
            project directory name.
        output_dir (string): Path where the project directory will be created.

    Return:
        string: Path where the project has been created.
    """
    logger = logging.getLogger("optimus")

    logger.debug("Project name: {}".format(project_name))
    logger.debug("Destination: {}".format(output_dir))

    try:
        # Just create project, disable input and just configure it from given context
        created = cookiecutter(
            template,
            extra_context={
                "project_name": project_name,
            },
            no_input=True,
            output_dir=output_dir,
        )
    # NOTE: This is actually the only exception from cookiecutter catched to emit a
    # critical log, however it got some others than may be raised. Eventually we will
    # catch a little more or catch cookiecutter base exception.
    except RepositoryNotFound as e:
        logger.critical(str(e))
        raise e

    logger.info("Created project to: {}".format(created))

    return created



# -*- coding: utf-8 -*-
"""
Module loader helpers
*********************

"""
import importlib
import logging
import sys

from optimus.conf.model import SettingsModel


def import_project_module(
    name,
    basedir=None,
    finding_module_err="Unable to find module: {0}",
    import_module_err="Unable to load module: {0}",
):
    """
    Load given module name.

    This is the new way technic, project base directory have to be loaded in
    ``sys.path`` with ``setup_project.setup_project`` before using it.

    NOTE:
        * This keeps deprecated "basedir" arg until migration ends and cleaning.

    Note that if you use this function to import successively the same module path,
    you may need to reload importation with something like this: ::

        mod = import_project_module(name)
        mod = importlib.reload(mod)

    Else you have unexpected behaviors like the second module returning content
    from a previously imported similar path module. However, this is a particular case
    that you may not encounter, this is mostly useful inside unittesting.

    Arguments:
        name (str): Module name to retrieve and import.

    Keyword Arguments:
        basedir (str): Base directory from where to find module name. If no
            base directory is given ``os.getcwd()`` is used. Default is
            ``None``. DEPRECATED: Still there temporary for compatible signature
            but the basedir is only used from "setup_project".
        finding_module_err (str): Message to output when the given module name
            is not reachable from ``basedir``.
        import_module_err (str): Message to output when the given module name
            raise exception when loaded.

    Returns:
        object: Finded and loaded module.
    """
    logger = logging.getLogger("optimus")
    logger.info('Loading "%s" module', name)

    # NOTE: Maybe we should raise better exception (with logged msg?), have to
    #       check
    # Try to locate module
    if importlib.util.find_spec(name):
        # Try to import module
        try:
            mod = importlib.import_module(name)
        # Module is invalid or unfound. Break, log and print out on any exception
        # during importation
        except Exception as error:
            logger.critical(import_module_err.format(name))
            # Print out useful exception
            raise error
            # sys.exit()
    # Unable to locate module, it's a critical failure
    else:
        logger.critical(finding_module_err.format(name))
        # TODO: We must not use sys.exit and raise a clear Exception instead,
        # higher layer level code will have to catch it and do the sys.exit/click.abort
        # itself if this is the required behavior. Actually sys.exit usage is opaque
        # msg = "Unable to found module: {}"
        # raise ImportError(msg.format(name))
        sys.exit()

    return mod


def import_settings_module(name, basedir=None):
    """
    Shortcut to have specific error message when loading settings module

    Arguments:
        name (str): Module name to retrieve from ``basedir``. This is Python path to
            the module from project base directory as loaded from project setup.

    Keyword Arguments:
        basedir (str): Base directory from where to find module name. If no
            base directory is given ``os.getcwd()`` is used. Default is
            ``None``.

    Returns:
        object: Finded and loaded module.
    """
    msg_finding = "Unable to find settings module: {0}"
    msg_import = "Unable to load settings module, it probably have errors: {0}"

    return import_project_module(
        name,
        basedir=basedir,
        finding_module_err=msg_finding,
        import_module_err=msg_import,
    )


def import_pages_module(name, basedir=None):
    """
    Shortcut to have specific error message when loading a page module

    Arguments:
        name (str): Module name to retrieve from ``basedir``. This is Python path to
            the module from project base directory as loaded from project setup.

    Keyword Arguments:
        basedir (str): Base directory from where to find module name. If no
            base directory is given ``os.getcwd()`` is used. Default is
            ``None``.

    Returns:
        object: Finded and loaded module.
    """
    msg_finding = "Unable to find pages module: {0}"
    msg_import = "Unable to load pages module, it probably have errors: {0}"
    return import_project_module(
        name,
        basedir=basedir,
        finding_module_err=msg_finding,
        import_module_err=msg_import,
    )


def load_settings(settings_module):
    """
    Load settings module.

    A shortcut to validate required settings are set then fill some missing settings to
    a default value.

    Arguments:
        settings_module (object): A settings module to load in model.

    Returns:
        optimus.conf.model.SettingsModel: Settings module validated and filled with
        defaults.
    """
    _settings = SettingsModel()
    _settings.load_from_module(settings_module)

    return _settings


def import_settings(name, basedir=None):
    """
    Import and load settings module.

    Arguments:
        name (string): Module name to retrieve from ``basedir``. This is Python path to
            the module from project base directory as loaded from project setup.

    Keyword Arguments:
        basedir (string): Base directory from where to find module name. If no
            base directory is given ``os.getcwd()`` is used. Default is
            ``None``.

    Returns:
        object: Settings module validated and filled with defaults.
    """
    settings_module = import_settings_module(name, basedir)

    return load_settings(settings_module)



from __future__ import absolute_import
from webassets.filter import Filter


__all__ = ("RCSSMin",)


class RCSSMin(Filter):
    """
    Minifies CSS using 'rcssmin' library.

    Requires the ``rcssmin`` package (https://github.com/ndparker/rcssmin).

    This is a simple webassets filter that has been merged but not
    released yet, so we ship it until new package release since 'cssmin' is
    totally outdated and doesn't work with Python 3.
    """

    name = "rcssmin"

    def output(self, _in, out, **kw):
        import rcssmin

        out.write(rcssmin.cssmin(_in.read()))



# -*- coding: utf-8 -*-
"""
I18n management
***************

I18n management support for Optimus environnment.

Only "messages.*" files for POT and PO files are managed and no other catalog
type.

TODO:
    * Use exceptions instead of logger.error;
    * Avoid to directly use object attributes, prefer to give needed vars as
      method args;
"""
import datetime
import io
import logging
import os
import shutil
import tempfile

from babel import Locale
from babel.util import LOCALTZ
from babel.messages.extract import extract_from_dir
from babel.messages.catalog import Catalog
from babel.messages.pofile import read_po, write_po
from babel.messages.mofile import write_mo


class I18NManager:
    """
    I18n manager for translation catalogs

    Made to work simply within Optimus environnment, so not all of babel
    options are used. This way the manager can work cleanly and is more easy
    to use.

    Arguments:
        settings (conf.model.SettingsModel): Settings registry instance.

    Attributes:
        catalog_name (string): Catalog filename template.
        catalog_path (string): Catalog language directory template.
        header_comment (string): Header comment to prepend to catalog files.
        settings (conf.model.SettingsModel): Settings registry instance.
        logger (logging.Logger): Optimus logger.
    """

    catalog_name = "messages.{0}"
    catalog_path = "{0}/LC_MESSAGES"
    header_comment = (
        "# Translations template for PROJECT project\n# Created " "by Optimus"
    )

    def __init__(self, settings):
        self.settings = settings
        self._pot = None
        self.logger = logging.getLogger("optimus")

    def get_template_path(self):
        """
        Return the full path to the catalog template file

        Returns:
            string: Catalog template file path.
        """
        return os.path.join(self.settings.LOCALES_DIR, self.catalog_name.format("pot"))

    def get_catalog_dir(self, locale):
        """
        Return the full path to a translations catalog directory

        Arguments:
            locale (string): Language identifier.

        Returns:
            string: Catalog directory path.
        """
        return os.path.join(self.settings.LOCALES_DIR, self.catalog_path.format(locale))

    def get_po_filepath(self, locale):
        """
        Return the full path to a translations catalog file

        Arguments:
            locale (string): Language identifier.

        Returns:
            string: Catalog file path.
        """
        return os.path.join(
            self.get_catalog_dir(locale), self.catalog_name.format("po")
        )

    def get_mo_filepath(self, locale):
        """
        Return the full path to a compiled translations catalog file

        Arguments:
            locale (string): Language identifier.

        Returns:
            string: Compiled catalog file path.
        """
        return os.path.join(
            self.get_catalog_dir(locale), self.catalog_name.format("mo")
        )

    def check_locales_dir(self):
        """
        Check if LOCALES_DIR directory exists

        Returns:
            boolean: True if base catalog directory exists.
        """
        return os.path.exists(self.settings.LOCALES_DIR)

    def check_template_path(self):
        """
        Check if the catalog template exists

        Returns:
            boolean: True if catalog template file exists.
        """
        return os.path.exists(self.get_template_path())

    def check_catalog_path(self, locale):
        """
        Check if a translations catalog exists

        Arguments:
            locale (string): Language identifier.

        Returns:
            boolean: True if catalog file exists.
        """
        return os.path.exists(self.get_po_filepath(locale))

    def parse_languages(self, languages):
        """
        Allways return a list of locale name from languages even if items are
        simple string or tuples. If tuple, assume its first item is the locale
        name to use.

        Arguments:
            languages (list): List of languages identifiers.

        Returns:
            dict: Dictionnary of languages identifiers.
        """
        _f = (
            lambda x: x[0] if isinstance(x, list) or isinstance(x, tuple) else x
        )  # noqa
        return map(_f, languages)

    def init_locales_dir(self):
        """
        Create catalog base directory defined from ``LOCALES_DIR`` settings
        if it does not allready exists.
        """
        if not self.check_locales_dir():
            self.logger.warning(("Locale directory does not exists, " "creating it"))
            os.makedirs(self.settings.LOCALES_DIR)

    def build_pot(self, force=False):
        """
        Extract translation strings and create Portable Object Template (POT)
        from enabled source directories using defined extract rules.

        Note:
            May only work on internal '_pot' to return without touching
            'self._pot'.

        Keyword Arguments:
            force (boolean): Default behavior is to proceed only if POT file
                does not allready exists except if this argument is ``True``.

        Returns:
            babel.messages.catalog.Catalog: Catalog template object.
        """
        if force or not self.check_template_path():
            self.logger.info(
                ("Proceeding to extraction to update the " "template catalog (POT)")
            )
            self._pot = Catalog(
                project=self.settings.SITE_NAME, header_comment=self.header_comment
            )
            # Follow all paths to search for pattern to extract
            for extract_path in self.settings.I18N_EXTRACT_SOURCES:
                msg = "Searching for pattern to extract in : {0}"
                self.logger.debug(msg.format(extract_path))
                extracted = extract_from_dir(
                    dirname=extract_path,
                    method_map=self.settings.I18N_EXTRACT_MAP,
                    options_map=self.settings.I18N_EXTRACT_OPTIONS,
                )
                # Proceed to extract from given path
                for filename, lineno, message, comments, context in extracted:
                    filepath = os.path.normpath(
                        os.path.join(
                            os.path.basename(self.settings.SOURCES_DIR), filename
                        )
                    )
                    self._pot.add(
                        message,
                        None,
                        [(filepath, lineno)],
                        auto_comments=comments,
                        context=context,
                    )

            with io.open(self.get_template_path(), "wb") as fp:
                write_po(fp, self._pot)

        return self._pot

    @property
    def pot(self):
        """
        Return the catalog template

        Get it from memory if allready opened, if allready exists then open
        it, else extract it and create it.

        Returns:
            babel.messages.catalog.Catalog: Catalog template object.
        """
        if self._pot is not None:
            return self._pot
        if self.check_template_path():
            with io.open(self.get_template_path(), "rb") as fp:
                self._pot = read_po(fp)
            return self._pot
        return self.build_pot()

    @pot.setter
    def pot(self, value):
        self._pot = value

    @pot.deleter
    def pot(self):
        del self._pot

    def safe_write_po(self, catalog, filepath, **kwargs):
        """
        Safely write or overwrite a PO(T) file.

        Try to write catalog to a temporary file then move it to its final
        destination only writing operation did not fail. This way initial file
        is not overwrited when operation has failed.

        Original code comes from ``babel.messages.frontend``.

        Arguments:
            catalog (babel.messages.catalog.Catalog): Catalog object to write.
            filepath (string): Catalog file path destination.
            **kwargs: Additional arbitrary keyword argumentsto pass to
                ``write_po()`` babel function.

        Returns:
            babel.messages.catalog.Catalog: Catalog template object.
        """
        tmpname = os.path.join(
            os.path.dirname(filepath),
            tempfile.gettempprefix() + os.path.basename(filepath),
        )
        # Attempt to write new file to a temp file, clean temp file if it fails
        try:
            with io.open(tmpname, "wb") as tmpfile:
                write_po(tmpfile, catalog, **kwargs)
        except:  # noqa
            os.remove(tmpname)
            raise

        # Finally overwrite file if previous job has succeeded
        try:
            os.rename(tmpname, filepath)
        except OSError:
            # We're probably on Windows, which doesn't support atomic
            # renames, at least not through Python
            # If the error is in fact due to a permissions problem, that
            # same error is going to be raised from one of the following
            # operations
            os.remove(filepath)
            shutil.copy(tmpname, filepath)
            os.remove(tmpname)

    def clone_pot(self):
        """
        Helper to clone POT catalog from writed file (not the one in memory)
        without to touch to ``_pot`` attribute.

        Returns:
            babel.messages.catalog.Catalog: Clone catalog template object.
        """
        self.logger.debug("Opening template catalog (POT)")
        with io.open(self.get_template_path(), "rb") as fp:
            catalog = read_po(fp)

        return catalog

    def init_catalogs(self, languages=None):
        """
        Create PO catalogs from POT if they dont allready exists

        Keyword Arguments:
            languages (list): List of languages to process. Default is
                ``None`` so languages are taken from ``LANGUAGES`` settings.

        Returns:
            list: List of language identifiers for created catalogs.
        """
        catalog_template = self.clone_pot()
        languages = self.parse_languages(languages or self.settings.LANGUAGES)
        created = []

        for locale in languages:
            translation_dir = self.get_catalog_dir(locale)
            catalog_path = self.get_po_filepath(locale)

            if not self.check_catalog_path(locale):
                msg = "Init catalog (PO) for language '{0}' to {1}"
                self.logger.debug(msg.format(locale, catalog_path))
                # write po from POT
                catalog_template.locale = Locale.parse(locale)
                catalog_template.revision_date = datetime.datetime.now(LOCALTZ)
                catalog_template.fuzzy = False

                if not os.path.exists(translation_dir):
                    os.makedirs(translation_dir)

                with io.open(catalog_path, "wb") as fp:
                    write_po(fp, catalog_template)

                created.append(locale)

        return created

    def update_catalogs(self, languages=None):
        """
        Update PO catalogs from POT

        Keyword Arguments:
            languages (list): List of languages to process. Default is
                ``None`` so languages are taken from ``LANGUAGES`` settings.

        Returns:
            list: List of language identifiers for updated catalogs.
        """
        languages = self.parse_languages(languages or self.settings.LANGUAGES)
        updated = []

        for locale in languages:
            catalog_path = self.get_po_filepath(locale)
            msg = "Updating catalog (PO) for language '{0}' to {1}"
            self.logger.info(msg.format(locale, catalog_path))

            # Open PO file
            with io.open(catalog_path) as fp:
                catalog = read_po(fp, locale=locale)

            # Update it from the template
            catalog.update(self.pot)
            self.safe_write_po(catalog, catalog_path)

            updated.append(locale)

        return updated

    def compile_catalogs(self, languages=None):
        """
        Compile PO catalogs to MO files

        Note:
            Errors have no test coverage since ``read_po()`` pass them through
            warnings print to stdout and this is not blocking or detectable.
            And so the code continue to the compile part.

        Keyword Arguments:
            languages (list): List of languages to process. Default is
                ``None`` so languages are taken from ``LANGUAGES`` settings.

        Returns:
            list: List of language identifiers for compiled catalogs.
        """
        languages = self.parse_languages(languages or self.settings.LANGUAGES)
        compiled = []

        for locale in languages:
            msg = "Compiling catalog (MO) for language '{0}' to {1}"
            self.logger.info(msg.format(locale, self.get_mo_filepath(locale)))
            with io.open(self.get_po_filepath(locale), "rb") as fp:
                #
                catalog = read_po(fp, locale)

            # Check errors in catalog
            errs = False
            for message, errors in catalog.check():
                for error in errors:
                    errs = True
                    self.logger.warning(
                        "Error at line {0}: {1}".format(message.lineno, error)
                    )
            # Don't overwrite previous MO file if there have been error
            if errs:
                self.logger.critical(
                    (
                        "There has been errors within the "
                        "catalog, compilation has been aborted"
                    )
                )
                break

            with io.open(self.get_mo_filepath(locale), "wb") as fp:
                write_mo(fp, catalog, use_fuzzy=False)

            compiled.append(locale)

        return compiled



# -*- coding: utf-8 -*-
import importlib
import logging
import os
import time

import click

from optimus.conf.loader import (
    import_pages_module,
    import_settings_module,
    load_settings,
)
from optimus.interfaces.build import builder_interface
from optimus.interfaces.watch import watcher_interface
from optimus.setup_project import setup_project
from optimus.utils import display_settings


@click.command("watch", short_help="Watch for changes in project sources")
@click.option(
    "--basedir",
    metavar="PATH",
    type=click.Path(exists=True),
    help=(
        "Base directory where to search for settings file. "
        "Default value use current directory."
    ),
    default=os.getcwd(),
)
@click.option(
    "--settings-name",
    metavar="NAME",
    help=(
        "Settings file name to use without '.py' extension. "
        "Default value is 'settings'."
    ),
    default="settings",
)
@click.pass_context
def watch_command(context, basedir, settings_name):
    """
    Watch for changes in project sources to automatically build project
    ressources
    """
    logger = logging.getLogger("optimus")

    # Set project before to be able to load its modules
    setup_project(basedir, settings_name)

    # Load current project settings and page map
    settings = import_settings_module(settings_name, basedir=basedir)
    # In test environment, force the module reload to avoid previous test cache to be
    # used (since the module have the same path).
    if context.obj["test_env"]:
        settings = importlib.reload(settings)

    settings = load_settings(settings)

    views = import_pages_module(settings.PAGES_MAP, basedir=basedir)
    if context.obj["test_env"]:
        views = importlib.reload(views)

    # Debug output
    display_settings(
        settings,
        ("DEBUG", "PROJECT_DIR", "SOURCES_DIR", "TEMPLATES_DIR", "LOCALES_DIR"),
    )

    logger.debug("Trigger pages build to start")
    build_env = builder_interface(settings, views)

    # Init and configure observer with events
    observer = watcher_interface(settings, views, build_env)

    logger.warning("Starting to watch sources, use CTRL+C to stop it")
    # Do not start observer during tests since we cannot manage interruption and
    # watcher threads
    if not context.obj["test_env"]:
        observer.start()
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            logger.warning("Stopping watcher..")
            observer.stop()
        observer.join()



"""
Logging
=======

Optimus makes usage of logging during its processes to inform user about
what it is doing or errors that occured.

To be more readable, its logger is configured to be colored using
``colorlog`` package.

"""
import logging

import colorlog


def set_loggers_level(names, level=logging.CRITICAL):
    """
    Set a log level on multiple loggers.

    Arguments:
        names (list): A list of logger names to set level.

    Keyword Arguments:
        level (integer): Logging level to set on all given logger names. Default to
            value from ``logging.CRITICAL``.
    """
    for item in names:
        logging.getLogger(item).setLevel(level)


def init_logger(level, printout=True):
    """
    Initialize app logger to configure its level/handler/formatter/etc..

    Arguments:
        level (str): Level name (``debug``, ``info``, etc..).

    Keyword Arguments:
        printout (bool): If False, logs will never be outputed.

    Returns:
        logging.Logger: Application logger.
    """
    root_logger = logging.getLogger("optimus")
    root_logger.setLevel(level)

    # Redirect outputs to the void space, mostly for usage within unittests
    if not printout:
        from io import StringIO

        dummystream = StringIO()
        handler = logging.StreamHandler(dummystream)
    # Standard output with colored messages
    else:
        handler = logging.StreamHandler()
        handler.setFormatter(
            colorlog.ColoredFormatter(
                "%(asctime)s - %(log_color)s%(message)s", datefmt="%H:%M:%S"
            )
        )

    root_logger.addHandler(handler)

    return root_logger



# -*- coding: utf-8 -*-
from optimus.i18n.manager import I18NManager


def po_interface(settings, init=False, update=False, compile_opt=False):
    """
    Manage project translation catalogs for all registred languages.

    You may enable all available modes. Modes are always processed in the same order:
    "init" then "update" and finally "compile".

    Arguments:
        settings (optimus.conf.model.SettingsModel): Settings object which define paths
            for locale directory and path for template sources to possibly scan.

    Keyword Arguments:
        init (boolean): Enable init mode to initialize POT file and "locale" directory.
        update (boolean): Enable update mode to refresh POT file and PO files for
            template changes.
        compile_opt (boolean): Enable compile mode to compile MO files from PO files.
    """
    # Proceed to operations
    i18n = I18NManager(settings)

    if init or update or compile_opt:
        i18n.init_locales_dir()
        i18n.build_pot(force=update)
        i18n.init_catalogs()

    if update:
        i18n.update_catalogs()

    if compile_opt:
        i18n.compile_catalogs()



# -*- coding: utf-8 -*-
import logging
import os
from string import ascii_letters, digits

import click

from cookiecutter.exceptions import RepositoryNotFound
from optimus.interfaces.starter import starter_interface
from optimus.starters import resolve_internal_template
from optimus.logs import set_loggers_level


@click.command("init", short_help="Create a new project from a cookiecutter template")
@click.argument("name")
@click.option(
    "--template",
    metavar="NAME",
    help=("A valid cookiecutter template for Optimus."),
    default="basic",
)
@click.option(
    "--destination",
    metavar="PATH",
    type=click.Path(exists=True),
    help=(
        "Directory path where to create the project directory. Default use the "
        "current directory"
    ),
    default=os.getcwd(),
)
@click.pass_context
def startproject_command(context, name, template, destination):
    """
    Create a new project from a template

    Expect one argument 'NAME' that will be the project name and its directory
    name. The name must be a valid Python module name.
    """
    # Mute all other loggers from cookiecutter and its dependancies
    set_loggers_level(["poyo", "cookiecutter", "binaryornot"])

    logger = logging.getLogger("optimus")

    # Valid that all characters from the name are : "_" character, letters,
    # identifier ::=  (letter|"_") (letter | digit | "_")*
    # This is not fully safe, user can create a project name using an installed
    # Python module that will override it and make some troubles in some case
    if name:
        if name[0] not in ascii_letters:
            logger.error("Project name must start with a letter")
            return
        for k in name[1:]:
            if k not in ascii_letters and k not in digits and k != "_":
                logger.error(
                    (
                        "Project name must only contains letters, "
                        "digits or '_' character"
                    )
                )
                return

    # Resolve possible internal template alias to a path
    template = resolve_internal_template(template)

    try:
        starter_interface(template, name, destination)
    except RepositoryNotFound:
        raise click.Abort()



# -*- coding: utf-8 -*-
import logging
import os
import shutil


def synchronize_assets_sources(from_path, to_path, src, dest, dry_run=False):
    """
    Copy (or replace existing) given source directory to destination.

    If source dir allready exists into given destination, existing dir is
    removed so given source dir replace it.

    Arguments:
        from_path (str): Base directory where to get the source dir to copy.
        to_path (str): Base directory where to put the source dir to copy.
        src (str): Source directory name (relative to ``from_path``) to copy.
        dest (str): NOTE: Unused, actually the source dir name is allways
            used as destination, dest should trigger a rename on copied item if
            given.

    Keyword Arguments:
        dry_run (bool): Enabled dry run mode, no directory will be created or
            removed. Default to ``False``.

    Returns:
        str: Copied destination path
    """
    logger = logging.getLogger("optimus")
    source = os.path.join(from_path, src)
    destination = os.path.join(to_path, src)

    if not os.path.exists(source):
        logger.warning(
            (
                "The given source does not exist and so can " "not be synchronized : {}"
            ).format(source)
        )
        return

    if os.path.exists(destination):
        logger.debug("Removing old asset destination: {}".format(destination))
        if not dry_run:
            shutil.rmtree(destination)

    logger.debug(
        ("Synchronizing asset from " '"{}" to "{}"').format(source, destination)
    )
    if not dry_run:
        shutil.copytree(source, destination)

    return destination



# -*- coding: utf-8 -*-
import logging

from webassets import Environment as AssetsEnvironment


class AssetRegistry(object):
    """
    Index all knowed files from registered bundles

    Attributes:
        map_dest_to_bundle (dict): Registry of asset paths associated to
            their asset bundle keyname.
        logger (logging.Logger): Optimus logger.
    """

    def __init__(self):
        self.map_dest_to_bundle = {}
        self.logger = logging.getLogger("optimus")

    def add_bundle(self, name, bundle):
        """
        Add a bundle to the registry

        Arguments:
            name (string): Bundle name as defined in the assets map.
            bundle (webassets.Bundle): Bundle to associate to given name.
        """
        # Little trick because a Bundle does not know its name in the
        # webassets environment
        bundle._internal_env_name = name

        for item in bundle.contents:
            self.map_dest_to_bundle[item] = name


def register_assets(settings):
    """
    Initialize webassets environment and its bundles.

    Arguments:
        settings (conf.model.SettingsModel): Settings registry instance.

    Returns:
        webassets.Environment: New configured Webasset environment.
    """
    logger = logging.getLogger("optimus")
    if not settings.ENABLED_BUNDLES:
        logger.warning(("Asset registering skipped as there are no enabled " "bundle"))
        return None
    logger.info("Starting asset registering")

    # Assets bundles
    AVAILABLE_BUNDLES = getattr(settings, "BUNDLES", {})

    # Initialize webassets environment
    assets_env = AssetsEnvironment()
    assets_env.debug = settings.DEBUG
    assets_env.url = settings.STATIC_URL
    assets_env.directory = settings.STATIC_DIR
    assets_env.load_path = [settings.SOURCES_DIR]
    assets_env.cache = settings.WEBASSETS_CACHE
    assets_env.url_expire = settings.WEBASSETS_URLEXPIRE

    #
    assets_env.optimus_registry = AssetRegistry()

    # Register enabled assets bundles
    for bundle_name in settings.ENABLED_BUNDLES:
        logger.debug("Registering bundle: {}".format(bundle_name))

        assets_env.register(bundle_name, AVAILABLE_BUNDLES[bundle_name])
        assets_env.optimus_registry.add_bundle(
            bundle_name, AVAILABLE_BUNDLES[bundle_name]
        )

    # When after bundle has been registered we can resolve it
    for bundle_name in settings.ENABLED_BUNDLES:
        logger.info("  Processing: {}".format(assets_env[bundle_name].resolve_output()))
        # Avoid to loop on every bundle part if we are not in debug logger
        if logger.getEffectiveLevel() == logging.DEBUG:
            for url in assets_env[bundle_name].urls():
                logger.debug("  - {}".format(url))

    return assets_env



# -*- coding: utf-8 -*-
"""
Project starters
================

Available shipped project templates to use with project starter.

"""
import os
import logging


STARTERS_PATH = os.path.normpath(
    os.path.dirname(__file__),
)

TEMPLATE_ALIAS = {
    "basic": os.path.join(STARTERS_PATH, "basic"),
}


def resolve_internal_template(path):
    """
    Resolve possible internal template alias to its path if it match.

    Arguments:
        path (string): An alias name defined in ``optimus.samples.TEMPLATE_ALIAS`` or
            a path or URL to a valid cookiecutter template for Optimus.

    Returns:
        string: Resolved internal template path if given name correspond to an alias. If
            it does not match any alias, just return the given name so it is used by
            cookiecutter to search itself for a path or an URL.
    """
    logger = logging.getLogger("optimus")

    if path in TEMPLATE_ALIAS:
        path = TEMPLATE_ALIAS[path]
        logger.debug("Resolved internal template path to: {}".format(path))

    return path



# -*- coding: utf-8 -*-
import logging
import io
import os

from jinja2 import Environment as Jinja2Environment
from jinja2 import FileSystemLoader
from webassets.ext.jinja2 import AssetsExtension

# Optional babel import
try:
    from babel import support as babel_support
except ImportError:
    babel_support = None

from optimus import __version__
from optimus.pages.registry import PageRegistry
from optimus.exceptions import ViewImproperlyConfigured


class PageBuilder(object):
    """
    Builder class to init Jinja2 environment and build given pages.

    Arguments:
        settings (conf.model.SettingsModel): Settings registry instance.

    Keyword Arguments:
        jinja_env (jinja2.Jinja2Environment): Jinja2 environment. Default is
            ``None``.
        assets_env (webassets.Environment): Webasset environment. Default is
            ``None``.
        dry_run (boolean): Enable dry run mode. Default is ``False``.

    Attributes:
        logger (logging.Logger): Optimus logger.
        settings (conf.model.SettingsModel): Settings registry instance.
        jinja_env (jinja2.Jinja2Environment): Jinja2 environment. Default is
            ``None``.
        assets_env (webassets.Environment): Webasset environment. Default is
            ``None``.
        internationalized (boolean): Indicate it internationalization is
            enabled. Will be automatically set to ``True`` if Jinja environment
            enable the i18n extension.
        translations (dict): Dictionnary of translation catalog indexed on
            language identifier.
        registry (optimus.pages.registry.PageRegistry): Registry of all knowed
            page from scanning.
        dry_run (boolean): Dry run mode.

    """

    def __init__(self, settings, jinja_env=None, assets_env=None, dry_run=False):
        self.logger = logging.getLogger("optimus")

        self.settings = settings

        self.assets_env = assets_env

        self.internationalized = False
        self.translations = {}

        self.jinja_env = jinja_env or self.get_environnement(assets_env)
        self.jinja_env.globals.update(self.get_globals())
        self.logger.debug("PageBuilder initialized")

        self.registry = PageRegistry()
        self.dry_run = dry_run

    def get_environnement(self, assets_env=None):
        """
        Init and configure Jinja environment.

        Automatically enable some extensions and link possible asset
        environment.

        Keyword Arguments:
            assets_env (webassets.Environment): Webasset environment. Default
                is ``None``. If empty, webassets will not be available from
                page templates.

        Returns:
            jinja2.Jinja2Environment: Configured Jinja2 environment.
        """
        exts = []
        self.logger.debug(
            ("No Jinja2 environment given, initializing a " "new environment")
        )

        # It the assets environment is given, active the Jinja extension to
        # use webassets
        if self.assets_env is not None:
            exts.append(AssetsExtension)

        # Enabled Jinja extensions
        for ext in self.settings.JINJA_EXTENSIONS:
            exts.append(ext)

        # Active i18n behaviors if i18n extension is loaded and Babel has been
        # imported
        if "jinja2.ext.i18n" in exts and babel_support is not None:
            self.internationalized = True
            self.logger.debug("'i18n' enabled")

        # Boot Jinja environment
        env = Jinja2Environment(
            loader=FileSystemLoader(self.settings.TEMPLATES_DIR), extensions=exts
        )

        # Enable Jinja filters
        for name, module in self.settings.JINJA_FILTERS.items():
            env.filters[name] = module

        if assets_env:
            env.assets_environment = assets_env

        return env

    def serialize_settings(self):
        """
        Get and return valid settings variables.

        Valid settings means only variable names full uppercase.

        Returns:
            dict: Settings variables.
        """
        items = {}

        for item in dir(self.settings):
            if not item.startswith("_") and item.isupper():
                items[item] = getattr(self.settings, item)

        return items

    def get_globals(self):
        """
        Get global context variables.

        Returns:
            dict: Shortcuts to some common settings like ``SITE`` options,
            static urls, Optimus version and finally all settings contained
            in ``_SETTINGS``.
        """
        domain_prefix = "http://{}"
        if self.settings.HTTPS_ENABLED:
            domain_prefix = "https://{}"

        return {
            "debug": self.settings.DEBUG,
            "SITE": {
                "name": self.settings.SITE_NAME,
                "domain": self.settings.SITE_DOMAIN,
                "web_url": domain_prefix.format(self.settings.SITE_DOMAIN),
            },
            "STATIC_URL": self.settings.STATIC_URL,
            "_SETTINGS": self.serialize_settings(),
            "OPTIMUS": __version__,
        }

    def get_translation_for_item(self, page_item):
        """
        Try to load the translations for the page language if any, then install
        it in Jinja2.

        It does not reload a language translations if a previous page has
        allready loaded it.

        Arguments:
            page_item (optimus.pages.views.PageViewBase): Page instance which
                its language identifier will be used to search for translation
                catalog.

        Returns:
            babel.support.Translations: Translations object to give to Jinja
            i18n extension.
        """
        if not babel_support:
            return None

        # Get page language object
        lang = page_item.get_lang()

        # Dont load again allredy registered catalog
        if lang.code not in self.translations:
            msg = " - Loading translations for locale: {} - {}"
            self.logger.debug(msg.format(lang.code, lang))
            self.translations[lang.code] = babel_support.Translations.load(
                self.settings.LOCALES_DIR, lang.code, "messages"
            )

        # Install it in the Jinja env
        self.jinja_env.install_gettext_translations(
            self.translations[lang.code], newstyle=False
        )

        return self.translations[lang.code]

    def scan_item(self, page_item):
        """
        Scan given page to retrieve template dependancies.

        Possibly connect settings to page instance if not allready done.

        Arguments:
            page_item (optimus.pages.views.PageViewBase): Page instance.

        Returns:
            string: All used templates from given page.
        """
        # Connect stored settings to page if not allready set
        try:
            page_item.settings
        except ViewImproperlyConfigured:
            page_item.settings = self.settings

        msg = " Scanning page: {}"
        self.logger.info(msg.format(page_item.get_destination()))

        return page_item.introspect(self.jinja_env)

    def scan_bulk(self, page_list):
        """
        Scan all given pages to set their dependancies

        TODO:
            Implement a 'settings' kwarg to pass to scan_item to connect
            settings object to view.

        Arguments:
            page_list (list): List of page instances.

        Returns:
            list: Every template name involved in scanned page instances.
        """
        self.logger.info("Starting page builds")

        if not page_list:
            self.logger.warning(
                ("Page scanning skipped as there are no " "registered pages")
            )
            return None

        knowed = set([])
        for page in page_list:
            found = self.scan_item(page)
            self.registry.add_page(page, found)
            knowed.update(found)

        return knowed

    def build_item(self, page_item):
        """
        Build given page.

        Possibly connect settings to page instance if not allready done.

        Arguments:
            page_item (optimus.pages.views.PageViewBase): Page instance.

        Returns:
            string: Destination path from builded page.
        """
        # Connect stored settings to page if not allready set
        try:
            page_item.settings
        except ViewImproperlyConfigured:
            page_item.settings = self.settings

        msg = " Building page: {}"
        self.logger.info(msg.format(page_item.get_destination()))

        # Optional i18n
        if self.internationalized:
            self.get_translation_for_item(page_item)

        # Template render
        content = page_item.render(self.jinja_env)

        destination_path = os.path.join(
            self.settings.PUBLISH_DIR, page_item.get_destination()
        )
        # Creating destination path if needed
        destination_dir, destination_file = os.path.split(destination_path)
        if not os.path.exists(destination_dir):
            msg = " - Creating new directory : {}"
            self.logger.debug(msg.format(destination_dir))
            if not self.dry_run:
                os.makedirs(destination_dir)
        # Write it
        self.logger.debug(" - Writing to: {}".format(destination_path))
        if not self.dry_run:
            with io.open(destination_path, "w") as fp:
                fp.write(content)

        return destination_path

    def build_bulk(self, page_list):
        """
        Build all given pages.

        Return all the effective builded pages

        Arguments:
            page_list (list): List of page instances.

        Returns:
            list: List of destination paths from builded pages.
        """
        self.logger.info("Starting page builds")

        if not page_list:
            self.logger.warning(
                ("Page management skipped as there are no " "registered pages")
            )
            return None

        builded = []
        for page in page_list:
            builded.append(self.build_item(page))

        return builded



"""
Interfaces
==========

An interface is the functional implementation for command line but usable from Python
code without tricks involved by command lines requirements like importing settings or
page modules.

"""



# -*- coding: utf-8 -*-
import click

from optimus import __version__ as optimus_version

from babel import __version__ as babel_version
from jinja2 import __version__ as jinja2_version
from webassets import __version__ as webassets_version
from watchdog.version import VERSION_STRING as watchdog_version

try:
    from cherrypy import __version__ as cherrypy_version
except ImportError:
    cherrypy_version = "Not installed"


@click.command("version", short_help="Print out versions informations")
@click.pass_context
def version_command(context):
    """
    Print out version information.
    """
    versions = (
        ("Optimus", optimus_version),
        ("Babel", babel_version),
        ("cherrypy", cherrypy_version),
        ("click", click.__version__),
        ("Jinja2", jinja2_version),
        ("watchdog", watchdog_version),
        ("webassets", ".".join([str(i) for i in webassets_version])),
    )

    for i, data in enumerate(versions, start=1):
        name, version = data
        # First version is Optimus displayed as root
        if i == 1:
            click.echo("{} {}".format(name, version))
        # Display termination character for the last version
        elif i == len(versions):
            click.echo("└── {} {}".format(name, version))
        # Every other versions
        else:
            click.echo("├── {} {}".format(name, version))



# -*- coding: utf-8 -*-
from optimus.exceptions import InvalidHostname


def get_host_parts(hostname, default_name="127.0.0.1", default_port=80):
    """
    Parse given hostname

    Arguments:
        hostname (str): A valid hostname containing a valid host name with
            optional port separated with a ``:``.

    Keyword Arguments:
        default_name (str): Default hostname when none is given. Default value
            is ``127.0.0.1``.
        default_port (int): Default port when none is given. Default value is
            ``80``.

    Returns:
        tuple: Host name (string) and port (integer).
    """
    name = None
    port = None
    hostparts = hostname.split(":")

    if len(hostparts) > 2:
        raise InvalidHostname("Invalid hostname format, too many ':'")
    elif len(hostparts) == 2:
        name, port = hostparts
        if not port or not name:
            raise InvalidHostname(
                ("Invalid hostname format, address or port " "is empty")
            )

        try:
            port = int(port)
        except ValueError:
            raise InvalidHostname("Invalid port given: {0}".format(port))
    else:
        name = hostparts[0]

    return (name or default_name, port or default_port)



# -*- coding: utf-8 -*-
import logging

from watchdog.events import PatternMatchingEventHandler
from watchdog.utils.patterns import match_any_paths

from optimus.watchers import BaseHandler


class AssetsWatchEventHandler(BaseHandler, PatternMatchingEventHandler):
    """
    Assets events handler

    Since assets filename change when they are rebuiled, this handler also
    performs page rebuild to include the right assets urls.

    Arguments:
        settings (optimus.conf.model.SettingsModel): Project settings.
        assets_env (webassets.Environment): Webasset environment.
        pages_builder (optimus.pages.builder.PageBuilder): Page builder object
            that is triggered to perform page building.
        args: Additional arguments to be passed to handler, commonly for
            watchdog API.

    Keyword Arguments:
        kwargs: Optionnal keyword arguments commonly for watchdog API.

    Attributes:
        settings (optimus.conf.model.SettingsModel): As given from arguments.
        assets_env (webassets.Environment): Webasset As given from arguments.
        pages_builder (optimus.pages.builder.PageBuilder): As given from
            arguments.
        logger (logging.Logger): Optimus logger.
    """

    def __init__(self, settings, assets_env, pages_builder, *args, **kwargs):
        self.settings = settings
        self.assets_env = assets_env
        self.pages_builder = pages_builder

        self.logger = logging.getLogger("optimus")

        super(AssetsWatchEventHandler, self).__init__(*args, **kwargs)

    def build_for_item(self, path):
        """
        Build bundle containing given asset path and all pages.

        Arguments:
            path (string): Asset path.

        Returns:
            list: List of builded pages.
        """
        rel_path = self.get_relative_asset_path(path)
        built = []
        self.logger.warning("--- Changes detected on: {} ---".format(rel_path))

        # Search in the registry if the file is a knowed asset from a bundle
        if rel_path in self.assets_env.optimus_registry.map_dest_to_bundle:
            bundle_name = self.assets_env.optimus_registry.map_dest_to_bundle[
                rel_path
            ]  # noqa

            msg = "Build required for bundle: {}"
            self.logger.debug(msg.format(bundle_name))

            # Trigger webassets update on bundle (wont do nothing when
            # 'settings.DEBUG' is True)
            urls = self.assets_env[bundle_name].urls()
            self.logger.debug("Rebuilded urls: %s", urls)

            # Launch all pages rebuild
            builds = self.pages_builder.build_bulk(
                self.pages_builder.registry.get_all_pages()
            )
            built.extend(builds)
        else:
            msg = "Path are not registered from any assets bundle: {}"
            self.logger.warning(msg.format(rel_path))

        return built

    def on_moved(self, event):
        """
        Called when a file or a directory is moved or renamed.

        Many editors don't directly change a file, instead they make a
        transitional file like ``*.part`` then move it to the final filename.

        Arguments:
            event: Watchdog event, either ``watchdog.events.DirMovedEvent`` or
                ``watchdog.events.FileModifiedEvent``.
        """
        # We are only interested for destination
        if match_any_paths(
            [event.dest_path],
            included_patterns=self.patterns,
            excluded_patterns=self.ignore_patterns,
            case_sensitive=self.case_sensitive,
        ):
            msg = "Change detected from a move on: {}"
            self.logger.debug(msg.format(event.dest_path))

            return self.build_for_item(event.dest_path)

        return []

    def on_created(self, event):
        """
        Called when a new file or directory is created.

        Arguments:
            event: Watchdog event, either ``watchdog.events.DirCreatedEvent``
                or ``watchdog.events.FileCreatedEvent``.
        """
        msg = "Change detected from a create on: {}"
        self.logger.debug(msg.format(event.src_path))

        return self.build_for_item(event.src_path)

    def on_modified(self, event):
        """
        Called when a file or directory is modified.

        Arguments:
            event: Watchdog event, ``watchdog.events.DirModifiedEvent`` or
                ``watchdog.events.FileModifiedEvent``.
        """
        msg = "Change detected from an edit on: {}"
        self.logger.debug(msg.format(event.src_path))

        return self.build_for_item(event.src_path)



# -*- coding: utf-8 -*-
import logging
import os

from optimus.utils.assets import synchronize_assets_sources


def init_directory(directory):
    """
    Shortcut to create given directory if it does not allready exists and
    output success to logger.

    Arguments:
        directory (str): Directory to create.

    Returns: ``True`` if directory has been created else ``False``.
    """
    logger = logging.getLogger("optimus")
    if not os.path.exists(directory):
        logger.debug("Creating directory: {}".format(directory))
        os.makedirs(directory)
        return True
    return False


def recursive_directories_create(project_directory, tree, dry_run=False):
    """
    Recursivly create directory structure from given tree.

    Arguments:
        project_directory (str): Directory where to create directory
            tree.
        tree (list): Directory tree to create, each item is either a
            string or a list. If an item is a string it is assumed to be a
            directory name to create. If an item is a list, it is assumed to be
            a list of sub directories to create.

            Sample tree: ::

                tree = [
                    [
                        'one',
                        [
                            [
                                'two'
                            ],
                            [
                                'three',
                                [
                                    ['foo'],
                                ],
                            ],
                        ]
                    ],
                ],

            Will turn to directory structure: ::

                one/
                ├── three
                │   └── foo
                └── two

    Keyword Arguments:
        dry_run (bool): Enabled dry run mode, no directory will be created.
            Default to ``False``.
    """
    logger = logging.getLogger("optimus")

    for item in tree:
        if len(item) > 0:
            new_dir = item[0]
            path_dir = os.path.join(project_directory, new_dir)
            if not os.path.exists(path_dir):
                logger.info("* Creating new directory : {}".format(path_dir))
                if not dry_run:
                    os.makedirs(path_dir)
            else:
                logger.warning(
                    ("* Following path allready exist : " "{}").format(path_dir)
                )
        # Follow children directories to create them
        if len(item) > 1:
            recursive_directories_create(path_dir, item[1], dry_run=dry_run)


def initialize(settings):
    """
    Init project directory structure.

    Arguments:
        settings (object): Settings object. Require setting attributes
            ``STATIC_DIR``, ``WEBASSETS_CACHE``, ``SOURCES_DIR``,
            ``FILES_TO_SYNC`` to be set.

    """
    init_directory(settings.STATIC_DIR)
    init_directory(settings.WEBASSETS_CACHE)

    if settings.FILES_TO_SYNC is not None:
        for item in settings.FILES_TO_SYNC:
            synchronize_assets_sources(
                settings.SOURCES_DIR, settings.STATIC_DIR, item, None
            )



# -*- coding: utf-8 -*-
from optimus.assets.registry import register_assets
from optimus.pages.builder import PageBuilder
from optimus.utils import initialize


def builder_interface(settings, views):
    """
    Build all enabled pages from given views module.

    Arguments:
        settings (optimus.conf.model.SettingsModel): Settings object which defines
            everything required for building.
        views (object): Module which defines page views to build, in fact the module
            object require only a ``PAGES`` attribute that is a list of Page view.

    Returns:
        dict: A dictionnary with initialized builder (``builder`` item), asset manager
        (``assets_env`` item) and the list of builded pages (``builded`` item).
    """
    # Initialize required structure according to settings
    initialize(settings)

    # Init asset manager
    assets_env = register_assets(settings)

    # Init page builder
    builder = PageBuilder(settings, assets_env=assets_env)

    # Proceed to page building from registered pages
    builded = builder.build_bulk(views.PAGES)

    return {
        "assets_env": assets_env,
        "builded": builded,
        "builder": builder,
    }



# -*- coding: utf-8 -*-
import logging


class PageRegistry(object):
    """
    Page registry

    Index templates and memorize page destination that use them.

    TODO:
        * elements should be named 'templates'.
        * get_pages_from_dependency() return a list of destinations using a
          template name, it should be named 'get_pages_from_template'
        * @map_dest_to_page should be named 'destinations_pages'

    Keyword Arguments:
        elements (dict): Initial element dictionnary. Default to an empty dict.

    Attributes:
        elements (string): Dictionnary indexed on template names which contain
            destinations using them.
        map_dest_to_page (string): Dictionnary indexed on destinations which
            contain their related page view.
        logger (logging.Logger): Optimus logger.
    """

    def __init__(self, elements={}):
        self.elements = {}
        self.map_dest_to_page = {}
        self.logger = logging.getLogger("optimus")

    def add_page(self, page, templates):
        """
        Add a page to registry.

        Arguments:
            page (optimus.pages.views.PageViewBase): Page instance
            templates (list): List of templates names to link to given page
                instance.
        """
        self.map_dest_to_page[page.get_destination()] = page

        for k in templates:
            if k in self.elements:
                self.elements[k].add(page.get_destination())
            else:
                self.elements[k] = set([page.get_destination()])

    def get_pages_from_dependency(self, template_name):
        """
        Get page list depending from a template.

        This method is not safe out of the context of scanned pages, because
        it use an internal map builded from the scan use by the add_page
        method. In short, it will raise a KeyError exception for every
        destination that is unknowned from internal map.

        Arguments:
            template_name (string): Template name to search for.

        Returns:
            list: List of page instances depending from given template name.
        """
        if template_name not in self.elements:
            msg = "Given template name is not registered: {}"
            self.logger.warning(msg.format(template_name))
            return []

        destinations = self.elements[template_name]

        return [self.map_dest_to_page[item] for item in destinations]

    def get_all_destinations(self):
        """
        Return all registered destinations

        Returns:
            list: List of all page destinations.
        """
        return [dest for dest, page in self.map_dest_to_page.items()]

    def get_all_pages(self):
        """
        Return all registered pages

        Returns:
            list: List of all page instances.
        """
        return [page for dest, page in self.map_dest_to_page.items()]



"""
Main entrance to commandline actions
"""
import click

from optimus.cli.version import version_command
from optimus.cli.startproject import startproject_command
from optimus.cli.build import build_command
from optimus.cli.watch import watch_command
from optimus.cli.po import po_command
from optimus.cli.runserver import runserver_command
from optimus.logs import init_logger


# Help alias on '-h' argument
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

# Default logger conf
OPTIMUS_LOGGER_CONF = ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", None)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-v",
    "--verbose",
    type=click.IntRange(min=0, max=5),
    default=4,
    metavar="INTEGER",
    help=(
        "An integer between 0 and 5, where '0' make a totaly silent output and '5' "
        "set level to DEBUG (the most verbose level). Default to '4' (Info level)."
    ),
)
@click.option(
    "--test-env",
    is_flag=True,
    hidden=True,
    help=(
        "Just an option to run for test and enable some specific tricks to run "
        "correctly"
    ),
)
@click.pass_context
def cli_frontend(ctx, verbose, test_env):
    """
    Optimus is a static site builder using Jinja2, webassets and Babel.
    """
    printout = True
    if verbose == 0:
        verbose = 1
        printout = False

    # Verbosity is the inverse of logging levels
    levels = [item for item in OPTIMUS_LOGGER_CONF]
    levels.reverse()
    # Init the logger config
    root_logger = init_logger(levels[verbose], printout=printout)

    # Init the default context that will be passed to commands
    ctx.obj = {
        "verbosity": verbose,
        "logger": root_logger,
        "test_env": test_env,
    }


# Attach commands methods to the main grouper
cli_frontend.add_command(version_command, name="version")
cli_frontend.add_command(startproject_command, name="init")
cli_frontend.add_command(build_command, name="build")
cli_frontend.add_command(watch_command, name="watch")
cli_frontend.add_command(po_command, name="po")
cli_frontend.add_command(runserver_command, name="runserver")
